@@ -2,7 +2,7 @@
 # -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
 #   Catfish - a versatile file searching tool
 #   Copyright (C) 2007-2012 Christian Dywan <christian@twotoasts.de>
-#   Copyright (C) 2012-2015 Sean Davis <smd.seandavis@gmail.com>
+#   Copyright (C) 2012-2020 Sean Davis <bluesabre@xfce.org>
 #
 #   This program is free software: you can redistribute it and/or modify it
 #   under the terms of the GNU General Public License version 2, as published
@@ -14,13 +14,17 @@
 #   PURPOSE.  See the GNU General Public License for more details.
 #
 #   You should have received a copy of the GNU General Public License along
-#   with this program.  If not, see <http://www.gnu.org/licenses/>.
+#   with this program.  If not, see <https://www.gnu.org/licenses/>.
+
+# pylint: disable=C0413
+# pylint: disable=C0114
+# pylint: disable=C0116
 
 import logging
-logger = logging.getLogger('catfish_search')
 
 import io
 import os
+import re
 import signal
 import subprocess
 import time
@@ -28,30 +32,47 @@ from itertools import permutations
 
 from mimetypes import guess_type
 
-from sys import version_info
-python3 = version_info[0] > 2
-
-engine_count = 0
+import gi
+gi.require_version('GLib', '2.0')  # noqa
+from gi.repository import GLib
 
 try:
-    from zeitgeist.client import ZeitgeistDBusInterface
-    from zeitgeist.datamodel import Event, TimeRange
-    from zeitgeist import datamodel
-    iface = ZeitgeistDBusInterface()
-    zeitgeist_support = True
-except Exception:
-    zeitgeist_support = False
+    from zeitgeist.client import ZeitgeistDBusInterface  # pylint: disable=E0401
+    from zeitgeist.datamodel import Event, TimeRange  # pylint: disable=E0401
+    from zeitgeist import datamodel  # pylint: disable=E0401
+    IFACE = ZeitgeistDBusInterface()
+    ZEITGEIST_SUPPORT = True
+except ImportError:
+    ZEITGEIST_SUPPORT = False
 
-if subprocess.call(['which', 'locate']) == 0:
-    locate_support = True
+LOGGER = logging.getLogger('catfish_search')
+ENGINE_COUNT = 0
+
+FNULL = open(os.devnull, 'w')
+if subprocess.call(['which', 'locate'],  # pylint: disable=R1703
+                   stdout=FNULL, stderr=subprocess.STDOUT) == 0:
+    LOCATE_SUPPORT = True
 else:
-    locate_support = False
+    LOCATE_SUPPORT = False
+FNULL.close()
 
 
-def string_regex(keywords, path):
+def get_keyword_list(keywords):
+    keywords = keywords.replace(",", " ").strip().lower()
+    kwords = []
+    matches = re.findall(r'\"(.+?)\"', keywords)
+    for match in matches:
+        newmatch = match.replace(" ", "\0")
+        newmatch = newmatch.replace("\"", "")
+        keywords = keywords.replace("\"%s\"" % match, newmatch)
+    for keyword in keywords.split(" "):
+        kwords.append(keyword.replace("\0", " "))
+    return kwords
+
+
+def string_regex(keywords, path):  # pylint: disable=W0613
     """Returns a string with the regular expression containing all combinations
     of the keywords."""
-    keywords = keywords.split()
     if len(keywords) == 0:
         return ''
     if len(keywords) == 1:
@@ -78,7 +99,8 @@ class CatfishSearchEngine:
     """CatfishSearchEngine is the collection of search backends that are used
     to perform a query.  Each backend is a CatfishSearchMethod"""
 
-    def __init__(self, methods=['zeitgeist', 'locate', 'walk']):
+    def __init__(self, methods=['zeitgeist', 'locate', 'walk'],
+                 exclude_paths=[]):
         """Initialize the CatfishSearchEngine.  Provide a list of methods to
         be included in the search backends.  Available backends include:
 
@@ -87,37 +109,43 @@ class CatfishSearchEngine:
          walk             'os.walk' to search for files (like find).
          zeitgeist        Zeitgeist indexing service to search for files.
         """
-        global engine_count
-        engine_count += 1
-        self.engine_id = engine_count
-        logger.debug(
-            "[%i] engine initializing with methods: %s", self.engine_id, str(methods))
+        self.stop_time = 0
+        self.keywords = ""
+
+        global ENGINE_COUNT
+        ENGINE_COUNT += 1
+        self.engine_id = ENGINE_COUNT
+        LOGGER.debug(
+            "[%i] engine initializing with methods: %s",
+            self.engine_id, str(methods))
         self.methods = []
         if 'zeitgeist' in methods:
-            if zeitgeist_support:
+            if ZEITGEIST_SUPPORT:
                 self.add_method(CatfishSearchMethod_Zeitgeist)
         if 'locate' in methods:
-            if locate_support:
+            if LOCATE_SUPPORT:
                 self.add_method(CatfishSearchMethod_Locate)
         if 'fulltext' in methods:
             self.add_method(CatfishSearchMethod_Fulltext)
         if 'walk' in methods:
             self.add_method(CatfishSearchMethod_Walk)
+        self.exclude_paths = exclude_paths
         initialized = []
         for method in self.methods:
             initialized.append(method.method_name)
-        logger.debug(
-            "[%i] engine initialized with methods: %s", self.engine_id, str(initialized))
+        LOGGER.debug(
+            "[%i] engine initialized with methods: %s",
+            self.engine_id, str(initialized))
         self.start_time = 0.0
 
     def __del__(self):
-        logger.debug("[%i] engine destroyed", self.engine_id)
+        LOGGER.debug("[%i] engine destroyed", self.engine_id)
 
     def add_method(self, method_class):
         """Add a CatfishSearchMethod the the engine's search backends."""
         self.methods.append(method_class())
 
-    def run(self, keywords, path, limit=-1, regex=False):
+    def run(self, keywords, path, limit=-1, regex=False):  # noqa
         """Run the CatfishSearchEngine.
 
         Each method is run sequentially in the order they are added.
@@ -126,35 +154,56 @@ class CatfishSearchEngine:
         filename is yielded if it matches the query.  False is also yielded
         afterwards to guarantee the interface does not lock up."""
         self.start_time = time.time()
-        keywords = keywords.replace(',', ' ').strip().lower()
+        self.stop_time = 0
 
-        logger.debug("[%i] path: %s, keywords: %s, limit: %i, regex: %s",
-                     self.engine_id, str(path), str(keywords), limit, str(regex))
+        keywords = get_keyword_list(keywords)
+        self.keywords = " ".join(keywords)
 
-        self.keywords = keywords
+        LOGGER.debug("[%i] path: %s, keywords: %s, limit: %i, regex: %s",
+                     self.engine_id, str(path), str(keywords), limit,
+                     str(regex))
 
         wildcard_chunks = []
-        for key in self.keywords.split():
+        keys = []
+        for key in keywords:
             if '*' in key:
                 wildcard_chunks.append(key.split('*'))
-
-        keywords = keywords.replace('*', ' ')
+            else:
+                keys.append(key)
 
         # For simplicity, make sure the path contains a trailing '/'
         if not path.endswith('/'):
             path += '/'
 
-        # Transform the keywords into a clean list.
-        keys = []
-        for key in keywords.split():
-            keys.append(key.strip())
+        # Path exclusions for efficiency
+        exclude = []
+        maybe_exclude = self.exclude_paths
+        for maybe_path in maybe_exclude:
+            if not path.startswith(maybe_path):
+                exclude.append(maybe_path)
 
         file_count = 0
         for method in self.methods:
-            logger.debug(
-                "[%i] Starting search method: %s", self.engine_id, method.method_name)
-            for filename in method.run(keywords, path, regex):
+            if self.stop_time > 0:
+                LOGGER.debug("Engine is stopped")
+                return
+            LOGGER.debug(
+                "[%i] Starting search method: %s",
+                self.engine_id, method.method_name)
+            for filename in method.run(keywords, path, regex,
+                                       self.exclude_paths):
                 if isinstance(filename, str) and path in filename:
+                    found_bad = False
+                    for filepath in exclude:
+                        if filename.startswith(filepath):
+                            if self.stop_time > 0:
+                                LOGGER.debug("Engine is stopped")
+                                return
+                            found_bad = True
+                    if found_bad:
+                        yield True
+                        continue
+
                     if method.method_name == 'fulltext' or  \
                             all(key in
                                 os.path.basename(filename).lower()
@@ -216,7 +265,7 @@ class CatfishSearchEngine:
             method.stop()
         self.stop_time = time.time()
         clock = self.stop_time - self.start_time
-        logger.debug("[%i] Last query: %f seconds", self.engine_id, clock)
+        LOGGER.debug("[%i] Last query: %f seconds", self.engine_id, clock)
 
 
 class CatfishSearchMethod:
@@ -228,7 +277,7 @@ class CatfishSearchMethod:
         """Base CatfishSearchMethod Initializer."""
         self.method_name = method_name
 
-    def run(self, keywords, path, regex=False):
+    def run(self, keywords, path, regex=False, exclude_paths=[]):  # pylint: disable=W0613
         """Base CatfishSearchMethod run method."""
         return NotImplemented
 
@@ -250,21 +299,114 @@ class CatfishSearchMethod_Walk(CatfishSearchMethod):
     def __init__(self):
         """Initialize the 'walk' Search Method."""
         CatfishSearchMethod.__init__(self, "walk")
+        self.running = False
 
-    def run(self, keywords, path, regex=False):
+    def get_dir_list(self, root, dirs, xdg_list,
+                     exclude_list, processed_links):
+        dirs = sorted(dirs, key=lambda s: s.lower())
+
+        # Prioritize: XDG, Visible (Linked), Dotfile (Linked)
+        xdgdirs = []
+        dotdirs = []
+        dotlinks = []
+        notdotdirs = []
+        notdotlinks = []
+
+        realroot = os.path.realpath(root)
+
+        for path in dirs:
+            path = os.path.join(root, path)
+            # Remove trailing slashes to ensure that calling os.path.basename()
+            # will not cut the path to an empty string
+            path = os.path.normpath(path)
+            if path in exclude_list:
+                continue
+            if path in xdg_list:
+                xdgdirs.append(path)
+                continue
+            islink = os.path.islink(path)
+            if islink:
+                realpath = os.path.realpath(path)
+                if realpath in processed_links:
+                    continue
+                # Respect the user's settings
+                if realpath in exclude_list:
+                    continue
+                # Sandbox search results to the starting path, don't allow:
+                # Start: /home/username/
+                # Link: ~/.wine/dosdevices/z:/ -> /
+                if realroot.startswith(realpath):
+                    continue
+                processed_links.append(realpath)
+            if os.path.basename(path).startswith("."):
+                if islink:
+                    dotlinks.append(path)
+                else:
+                    dotdirs.append(path)
+            else:
+                if islink:
+                    notdotlinks.append(path)
+                else:
+                    notdotdirs.append(path)
+
+        dirlist = xdgdirs + notdotdirs + notdotlinks + dotdirs + dotlinks
+        return dirlist
+
+    def run(self, keywords, path, regex=False, exclude_paths=[]):
         """Run the search method using keywords and path.  regex is not used
         by this search method.
 
         This function is a generator and will yield files as they are found or
         True if still running."""
+        exclude = []
+        maybe_exclude = exclude_paths
+        for maybe_path in maybe_exclude:
+            if not path.startswith(maybe_path):
+                exclude.append(maybe_path)
+
         self.running = True
         if isinstance(keywords, str):
             keywords = keywords.replace(',', ' ').strip().split()
-        for root, dirs, files in os.walk(path, False):
+
+        # Enable symbolic link directories, but process once
+        processed_links = []
+
+        # Grab the special directory list to get them precedence
+        xdgdirlist = [
+            GLib.get_user_special_dir(GLib.USER_DIRECTORY_DESKTOP),
+            GLib.get_user_special_dir(GLib.USER_DIRECTORY_DOCUMENTS),
+            GLib.get_user_special_dir(GLib.USER_DIRECTORY_DOWNLOAD),
+            GLib.get_user_special_dir(GLib.USER_DIRECTORY_MUSIC),
+            GLib.get_user_special_dir(GLib.USER_DIRECTORY_PICTURES),
+            GLib.get_user_special_dir(
+                GLib.USER_DIRECTORY_PUBLIC_SHARE),
+            GLib.get_user_special_dir(GLib.USER_DIRECTORY_TEMPLATES),
+            GLib.get_user_special_dir(GLib.USER_DIRECTORY_VIDEOS),
+        ]
+
+        for root, dirs, files in os.walk(top=path, topdown=True,
+                                         onerror=None,
+                                         followlinks=True):
+            # Bail once the search has been canceled
             if not self.running:
                 break
-            paths = dirs + files
-            paths.sort()
+
+            # Check if we've already processed symbolic paths
+            if os.path.islink(root):
+                realpath = os.path.realpath(root)
+                if realpath in processed_links:
+                    yield True
+                    continue
+                processed_links.append(realpath)
+
+            # Prioritize and drop excluded paths
+            dirs[:] = self.get_dir_list(
+                root, dirs, xdgdirlist, exclude, processed_links)
+
+            paths = sorted(dirs + files)
+
+            # Check paths in the second and deeper levels of the selected
+            # directory
             for path in paths:
                 if any(keyword in path.lower() for keyword in keywords):
                     yield os.path.join(root, path)
@@ -293,7 +435,7 @@ class CatfishSearchMethod_Fulltext(CatfishSearchMethod):
         self.running = False
         self.exact = False
 
-    def run(self, keywords, path, regex=False):
+    def run(self, keywords, path, regex=False, exclude_paths=[]):  # noqa
         """Run the search method using keywords and path.  regex is not used
         by this search method.
 
@@ -312,7 +454,7 @@ class CatfishSearchMethod_Fulltext(CatfishSearchMethod):
                     find_keywords_backup.append(keyword)
 
         # Start walking the folder structure.
-        for root, dirs, files in os.walk(path):
+        for root, dirs, files in os.walk(path):  # pylint: disable=W0612
             if self.force_stop:
                 break
 
@@ -335,12 +477,12 @@ class CatfishSearchMethod_Fulltext(CatfishSearchMethod):
                                     break
 
                                 if self.exact:
-                                    if keywords in line:
+                                    if " ".join(keywords) in line:
                                         yield os.path.join(root, filename)
                                         break
                                 else:
                                     if any(keyword in line.lower()
-                                            for keyword in keywords):
+                                           for keyword in keywords):
                                         found_keywords = []
                                         for find_keyword in find_keywords:
                                             if find_keyword in line.lower():
@@ -380,15 +522,16 @@ class CatfishSearchMethod_Zeitgeist(CatfishSearchMethod):
     def __init__(self):
         """Initialize the Zeitgeist SearchMethod."""
         CatfishSearchMethod.__init__(self, "zeitgeist")
+        self.stop_search = False
 
-    def run(self, keywords, path, regex=False):
+    def run(self, keywords, path, regex=False, exclude_paths=[]):
         """Run the Zeitgeist SearchMethod."""
         self.stop_search = False
         event_template = Event()
         time_range = TimeRange.from_seconds_ago(60 * 3600 * 24)
         # 60 days at most
 
-        results = iface.FindEvents(
+        results = IFACE.FindEvents(
             time_range,  # (min_timestamp, max_timestamp) in milliseconds
             [event_template, ],
             datamodel.StorageState.Any,
@@ -435,11 +578,11 @@ class CatfishSearchMethodExternal(CatfishSearchMethod):
         self.command = []
         self.process = None
 
-    def assemble_query(self, keywords, path):
+    def assemble_query(self, keywords, path):  # pylint: disable=W0613
         """Base assemble_query method."""
-        return None
+        return False
 
-    def run(self, keywords, path, regex=False):
+    def run(self, keywords, path, regex=False, exclude_paths=[]):
         """Run the search method using keywords and path.
 
         This function returns the process.stdout generator and will yield files
@@ -461,10 +604,10 @@ class CatfishSearchMethodExternal(CatfishSearchMethod):
     def process_output(self, output):
         """Return the output text."""
         if isinstance(output, io.BufferedReader):
-            return map(lambda s: s.decode(encoding='UTF8').strip(),
+            return map(lambda s: s.decode(encoding='UTF8',
+                                          errors='replace').strip(),
                        output.readlines())
-        else:
-            return output
+        return output
 
     def status(self):
         """Return the current search status."""
@@ -496,9 +639,32 @@ class CatfishSearchMethod_Locate(CatfishSearchMethodExternal):
     def __init__(self):
         """Initialize the Locate SearchMethod."""
         CatfishSearchMethodExternal.__init__(self, "locate")
-        self.command = ["locate", "-i", "%path*%keywords*", "--existing"]
+        self.caps = self.get_capabilities()
+        if self.caps["existing"]:
+            self.command = ["locate", "-i", "%path*%keywords*", "--existing"]
+        else:
+            self.command = ["locate", "-i", "%path*%keywords*"]
+
+    def get_capabilities(self):
+        caps = {
+            "existing": False,
+            "regex": False
+        }
+        try:
+            details = subprocess.check_output(["locate", "--help"])
+            details = details.decode("utf-8")
+            if "--existing" in details:
+                caps["existing"] = True
+            if "--regex" in details or "--regexp" in details:
+                caps["regex"] = True
+
+        except subprocess.CalledProcessError:
+            pass
+        return caps
 
     def assemble_query(self, keywords, path):
         """Assemble the search query."""
-        return ["locate", "--regex", "-i", "{}".format(string_regex(keywords,
-                                                                    path))]
+        if self.caps["regex"]:
+            return ["locate", "--regex", "--basename", "-i",
+                    "{}".format(string_regex(keywords, path))]
+        return ["locate", "-i", "%path*", str(keywords)]
